//package urdar;
import java.util.*;

/** The <code>BString</code> class represents the bit strings on which the organisms feed. 
 * Each bit string is characterised by its binary sequence, length and metabolic history (i.e.
 * the species which it has been metabolised by). 
 * * @author Philip Gerlee
 * @author Torbjorn Lundh
 * @version 0.1 1st Jan 2011
 */
public class BString {
	/** 
	 * The length of the string
	 */
	public int length;
	/** 
	 * The bit sequence
	 */
	public boolean[] S; //bit sequence
	/** 
	 * The metabolic history of the string
	 */
	public Vector<Integer> H; //metabolic history



	/** Creates a string from a given bit sequence
	 * @param S	 The bit sequence
	 */
	public BString(boolean[] S) {
		this.S = S;
		length = S.length;
		H = new Vector<Integer>();
		H.add(new Integer(0));	
	}

	/** Creates a string from a given bit sequence and metabolic history
	 * @param S The bit sequence
	 * @param H metabolic history
	 */
	public BString(boolean[] S, Vector<Integer> H) {
		this.S = S;
		length = S.length;
		this.H = H;
	}
	/** Creates a string with a given length of certain type with bitflips
	 * @param size String length
	 * @param E Bitflip probability
	 * @param mode Different string modes:<br> 
	 * 0 (default) = Bitflips on repeated 0-strings and 1-strings.<br>
	 * 1 = Bitflips on repeated ...111111... pattern. <br>
	 * 2 = Bitflips on repeated ...01010101... pattern. <br>
	 * 3 = Bitflips on repeated ...011011011... pattern. <br>
	 * 4 = Bitflips on repeated ...011101110111... pattern. <br>	 
	 * */
	public BString(int size, double E, int mode) { 
		this.length = size;
		S = new boolean[size];
		H = new Vector<Integer>();
		H.add(new Integer(0));
		switch (mode) {
		case 1:
			for (int i = 0; i < size; i++) {
				S[i] = Math.random() < E;
			}
			break;
		case 2:
			for (int i = 0; i < size; i++) {
				if (i%2==0) {
					S[i]=true;
				}
				else {
					S[i]=false;
				}
				if (Math.random() < E) {
					S[i] = !S[i];
				}
			}
			break;
		case 3:
			for (int i = 0; i < size; i++) {
				if (i%3==0) {
					S[i]=true;
				}
				if (i%3==1) {
					S[i]=false;
				}
				if (i%3==2) {
					S[i]=false;
				}
				if (Math.random() < E) { 
					S[i] = !S[i];
				}
			}
			break;
		case 4:
			for (int i = 0; i < size; i++) {
				if (i%4==0) {
					S[i]=true;
				}
				if (i%4==1) {
					S[i]=false;
				}
				if (i%4==2) {
					S[i]=false;
				}
				if (i%4==3) {
					S[i]=false;
				}
				if (Math.random() < E) {
					S[i] = !S[i];
				}
			}
			break;
		default:
			if (Math.random() < 0.5) {
				for (int i = 0; i < size; i++) {
					S[i] = Math.random() < E;
				}
			}
			else {
				for (int i = 0; i < size; i++) {
					S[i] = Math.random() > E;
				}
			}
		}

	}

	/** Returns bit at position i
	 * @param i position along the string
	 * @return bit at position i
	 */
	public boolean get(int i) {
		return S[i];
	}

	/** Returns metabolic history
	 * @return the metabolic history as a Vector of Integers representing the species that have metabolised the string
	 */
	public Vector<Integer> getHistory() {
		return H;	
	}

	/** Prints the string to Std.out
	 */
	public void printString() {
		for (int i = 0; i < length; i++) {
			if (S[i]) {
				System.out.print(1);
			}
			else {
				System.out.print(0);
			}
		}
		System.out.println();
	}

	/** Induce random bitflips, not really used
	 * @param rate probability of bitflip
	 */
	public void radiate(double rate) {
		for (int i = 0; i < length; i++) {
			if (Math.random() < rate) {
				S[i] = !S[i];
			}
		}
	}

	/**  Returns the fraction of ones in the string
	 * @return the fraction of 1s in the string
	 */
	public double getFraction() {
		double sum = 0.0;
		for (int i = 0; i < length; i++) {
			if (S[i]) {
				sum = sum + 1.0;
			}
		}
		return (double) sum/length;
	}

	/** Applies a Cellular Automaton rule to the string once
	 * @param rule the CA-rule applied to the string
	 */
	public void manipulate(int rule) {
		//int k;
		int n;
		int[] R = new int[8];
		boolean[] Sn = new boolean[length];
		H.add(new Integer(rule));
		for (int i = 0; i < 8; i++) { //Conversion of decimal rule to bitstring
			R[i] = rule - 2*(rule/2);
			rule = rule/2;
		}
		for (int i = 0; i < length; i++) { //Applying the CA rule
			if (i==0) {
				n = (S[length-1] ? 1 : 0)*4 + (S[i] ? 1 : 0)*2 + (S[(i+1)%length] ? 1: 0)*1;
				Sn[i]=R[n]==1;
			}
			else {
				n = (S[i-1] ? 1 : 0)*4 + (S[i] ? 1 : 0)*2 + (S[(i+1)%length] ? 1: 0)*1;
				Sn[i]=R[n]==1;
			}
		}
		S = Sn;
	}

	/** Applies a Cellular Automaton rule to the string once and returns the modified string
	 * @param rule the CA-rule applied to the string
	 * @return the modified string
	 */
	public BString change(int rule) {
		int n;
		int[] R = new int[8];
		boolean[] Sn = new boolean[length];
		H.add(new Integer(rule));
		//Converting the rule to a binary representation
		for (int i = 0; i < 8; i++) { 
			R[i] = rule - 2*(rule/2);
			rule = rule/2;
		}
		//Applying the CA-rule to the bit string
		for (int i = 0; i < length; i++) { 
			if (i==0) {
				n = (S[length-1] ? 1 : 0)*4 + (S[i] ? 1 : 0)*2 + (S[(i+1)%length] ? 1: 0)*1;
				Sn[i]=R[n]==1;
			}
			else {
				n = (S[i-1] ? 1 : 0)*4 + (S[i] ? 1 : 0)*2 + (S[(i+1)%length] ? 1: 0)*1;
				Sn[i]=R[n]==1;
			}
		}
		return new BString(Sn);
	}

	/** Deep-copy of the string
	 * @return copy of the string
	 */
	public BString cloneString() {
		boolean[] St = new boolean[length];
		Vector<Integer> Ht = new Vector<Integer>();
		for (int i = 0; i < length; i++) {
			St[i]=S[i];
		}
		for (int i = 0; i < H.size(); i++) {
			Ht.add(i,H.get(i));
		}
		return new BString(St,Ht);
	}

	/** Compares two strings
	 * @return true if the two strings have the same bit sequence, false otherwise
	 */
	public boolean equals(BString Si) {
		boolean same = false;
		int N = 0;
		for (int i = 0; i < length; i++) {
			if (S[i]==Si.get(i)) {
				N++;
			}
			else {
				break;
			}
		}
		if (N==length) {
			same = true;
		}
		return same;
	}

	/**Calculates the approximation of the entropy S of a string B
	 * @param level the level at which the entropy is calculated
	 * @return the approximate entropy, i.e. S(B) &asymp; S<sub>L+1</sub>(B) - S<sub>L</sub>(B), where S<sub>L</sub> is the block entropy of length L
	 */
	public double Entropy(int level) {
		double sum = 0.0;
		int ones = 0;
		//In this case the entropy is given by -Sum p_i*log p_i, where p_i = prob. to find zero/one in the string
		if (level == 1) { 
			for (int i = 0; i < length; i++) {
				ones = ones + (S[i] ? 1 : 0);
			}
			if (ones != length && ones != 0) {
				sum = (-(double) ones/length)*Math.log((double) ones/length) + (-(double) (length-ones)/length)*Math.log((double) (length-ones)/length);
			}
			else {
				sum = 0.0;
			}
			sum = sum/Math.log(2);
		}
		//otherwise the entropy is defined as the difference between block entropies
		else { 
			sum = blockEntropy(level+1) - blockEntropy(level);
		}
		return sum;
	}

	/**Calculates the block entropy of string B
	 * @param level the level at which the block entropy is calculated
	 * @return the block entropy S<sub>L</sub>(B) = - &Sigma; p(&sigma;<sub>n</sub>) log p(&sigma;<sub>n</sub>)
	 */
	public double blockEntropy(int level) {
		int[] pow = {1,2,4,8,16,32}; //Avoid calculation of powers of 2
		int sym = 0;
		double sum = 0.0;

		//There are pow[level]=2^level different blocks of size $level
		int[] P = new int[pow[level]]; //The probability array
		for (int i = 0; i < P.length; i++) {
			P[i] = 0;
		}
		int[] T = new int[level]; //temporary array which holds the blocks
		for (int i = 0; i < length; i++) {
			for (int j = 0; j < level; j++) {
				T[j] = S[(i+j)%length] ? 1 : 0;
			}
			sym = 0;
			for (int j = 0; j < level; j++) {
				//each block is mapped to its corresponding decimal number, i.e. 00 --> 0, 10-->1, 01-->2, 11-->3
				sym = sym + T[j]*pow[j];   
			}
			P[sym] = P[sym] + 1; //count the number of blocks of each kind
		}
		sum = 0.0;
		//calculate the entropy of this distribution
		for (int i = 0; i < P.length; i++) { 
			if (P[i] > 0) {
				sum = sum + (-(double) P[i]/length)*Math.log((double) P[i]/length);
			}
		}
		return sum/(Math.log(2)); //normalise to bits
	}

	/** Calculates difference in entropy between this string and and another string
	 * @param S0 the string with which the difference is calculated
	 * @param L the level of entropy calculation
	 * @return the entropy difference E(S) - E(S0) between the two strings
	 */
	public double deltaE(BString S0, int L) {
		return this.Entropy(L) - S0.Entropy(L);
	}

}

