//package urdar;

/** The <code>Organism</code> class represents the organisms in the simulation. 
 * Each organism is characterised by a number of features, the genotype or cellular automaton
 * rule the organism implements being the most important.
 * @author Philip Gerlee
 * @author Torbjorn Lundh
 * @version 0.1 1st Jan 2011
 */
public class Organism {
    /** 
     * The organism identity number (unique to each instance of the class)
     */
    public int OID;
    /** 
     * The organism genotype identity number. Not currently used, but makes it possible to track lineages
     */
    public int GID; 
    /** 
     * The age of the organism counted in updates
     */
    public int age;
    /** 
     * The CA rule implemented by the organism, i.e. its pheno/genotype
     */
    public int G; 
    /** 
     * The most recent entropy change achieved by the organism
     */
    public double dE;
    /** 
     * The bit string attached to the organism
     */
    public BString S; 
    /** 
     * A backup copy of the bit string attached to the organism
     */
    public BString S0; //copy of original string
    /** 
     * Shows if the organism was a mutant at birth
     */
    public boolean mutated;
    /** 
     * Shows if the organism is a newborn
     */
    public boolean isNew;
    /** 
     * The current highest organism ID
     */
    public static int currentOID = 1; 
    /** 
     * The current highest genotype ID
     */
    public static int currentGID = 1; 
    /** 
     * The copy mutation probability
     */
    public static double pc; 

    /** Creates an Organism with given properties
     * @param G the CA-rule/genotype of the organism
     * @param OID the organism ID 
     * @param GID the genotype ID
     */
    public Organism(int G, int OID, int GID) {
	this.G = G;
	this.OID = OID;
	this.GID = GID;
	age = 0;
	dE = 0.0;
	mutated = false;
	isNew = true;
    }

    /** Returns an OID and increments the counter
     * @return the next available organism ID
     */
    public static int getNextOID() { //Returns the next cell ID and increments
	currentOID++;
	return currentOID;
    }

    /** Returns a GID and increments the counter
     * @return the next available genotype ID
     */
    public static int getNextGID() { //Returns the next genotype ID and increments
	currentGID++;
	return currentGID;
    }

    /** Returns an OID
     * @return the highest current organism ID
     */    
    public static int getCurrentOID() { //Just return
	return currentOID;
    }

    /** Returns a GID 
     * @return the highest current genotype ID
     */
    public static int getCurrentGID() { //Just return
	return currentGID;
    }

    /** Sets the mutation rate
     * @param p the copy mutation rate
     */
    public static void setMutRate(double p) {
	pc = p;
    }

    /** Attaches a new bit string to the organism and creates a backup
     * @param Snew the string to be attached
     */
    public void setString(BString Snew) {
	S = Snew;
	S0 = S.cloneString();
    }

    /** Returns the OID 
     * @return the organism ID number
     */
    public int getOID() {
	return OID;
    }

    /** Sets the organism ID number
     * @param ID the organism ID number
     */
    public void setOID(int ID) {
	OID = ID;
    }

    /** Returns the organism genotype ID number 
     * @return the organism genotype ID
     */
    public int getGID() {
	return GID;
    }

    /** Sets the organism genotype ID number
     * @param ID the organism genotype ID number
     */
    public void setGID(int ID) {
	GID = ID;
    }

    /** Check if the organism is mutated
     * @return true if organism was mutated at birth
     */
    public boolean isMutated() {
	return mutated;
    }

    /** Mark the organism as mutated
     * @param b true if mutated
     */
    public void mutated(boolean b) {
	mutated = b;
    }

    /** Clear the isNew tag
     */
    public void clearTag() {
	isNew = false;
    }	    

    /** Get the organism genotype 
     * @return the organism genotype
     */
    public int getGenome() {
	return G;
    }

    /** Sets the organism genotype
     * @param G the organism genotype
     */
    public void setGenome(int G) {
	this.G = G;
    }

    /** Get the string attached to the organism
     * @return the string attached to the organism
     */
    public BString getString() {
	return S;
    }

    /** Get the backup string attached to the organism
     * @return the backup string attached to the organism
     */
    public BString getOrigString() {
	return S0;
    }

    /** Get the most recent entropy difference
     * @return the latest entropy difference
     */
    public double getLatestIO() {
	return dE;
    }

    /** Get the age of the organism
     * @return the age in updates
     */
    public int getAge() {
	return age;
    }

    /** Updates the organism 
     * @param L the level of entropy calculation
     * @return 0 - if dE < 0 and no division occurs,<br>1 - if dE > 0 and the organism is ready to divide,<br>2 - if dE > 0 but no division should occur,
     */
    public int update(int L) {
	int div = 0;
	//Make sure the organism is not new born
	if (!isNew) { 
	    double Et = 0;
	    double beta = -10.0;
	    S.manipulate(G); //Manipulate the current string
	    Et = S.deltaE(S0,L); //Calculate entropy change
	    //Division occurs with probability P(Et)
	    if (Et > 0) {
		dE = Et;
		if ((1-Math.exp(beta*Et))/(1-Math.exp(beta)) > Math.random()) {  
		    div = 1;
		}
		else {
		    div = 2;
		}
	    }
	    else {
		dE = 0;
	    }
	    age++;
	}
	return div;
    }
    
    /** Manipulates a test string and returns it
     * @param St the test string
     * @return a manipulated version of the input string
     */
    public BString test(BString St) {
	St.manipulate(G);
	return St;
    }

    /** Swaps the current string and creates a backup
     * @param W the World from which the new string is picked
     * @param Sx the string which is returned
     */
    public void IO(World W, BString Sx) {
	S = W.swapString(Sx);
	S0 = S.cloneString();
    }

    /** Creates a daughter organism under mutations. With probability <code>pc</code> the daughter 
     * is identical with the parent (except for <code>OID</code> ) and with complementary
     * probability 1-<code>pc</code> it is assigned a random genotype
     * @return the daughter organism
     */
    public Organism divide() {
	boolean mut = false;
	//Create new organism with the same genotype, GID, but new OID
	Organism child = new Organism(G,getNextOID(),getGID());
	if (Math.random() < pc) { //Mutation
	    child.setGenome((int) Math.floor(255*Math.random()));
	    mut = true;
	}
	if (mut) {
	    child.mutated(true);
	    child.setGID(getNextGID());
	}
	return child;
    }
}

